mod sync_dir;

use std::{env, fs, io, thread, time};
use std::collections::HashMap;
use std::fs::File;
use std::io::{Read, stdin, Write};
use std::path::{Path, PathBuf};
use std::sync::Arc;
use futures_lite::AsyncReadExt;
use tokio_stream::{Stream, StreamExt};
use anyhow;
use std::time::Duration;
use iroh::base::node_addr::AddrInfoOptions;
use iroh::client::Doc;
use iroh::client::docs::{LiveEvent, ShareMode};
use iroh::docs::store::Query;
use iroh::node::Node;
use tokio::sync::Mutex;
use std::sync::mpsc::{channel, TryRecvError};
use iroh::blobs::store::fs::Store;
use iroh::docs::{Author, AuthorId, ContentStatus};
use iroh::net::key;
use iroh::node;
use notify::{DebouncedEvent, RecursiveMode, watcher, Watcher};
use crate::sync_dir::SyncDir;

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let watch_rate = 100; // each 100ms program check if there is any changes in the sync directory
    let args: Vec<String> = env::args().skip(1).collect(); // agrs (sync directory and ticket)

    let base_path = env::current_dir().unwrap().join(args[0].parse::<String>().unwrap()); // sync dir
    println!("BASE_PATH: {:?}", base_path);
    let store_path = "./iroh_data/".to_string(); // place where to persist data from doc

    // creating SyncDir
    let sync_dir = if args.len() > 1 {
        SyncDir::new(base_path.clone(), store_path.clone(), Some(args[1].clone())).await?
    }else{
        SyncDir::new(base_path.clone(), store_path.clone(), None).await?
    };
    let sync_dir = Arc::new(Mutex::new(sync_dir)); // wrap it into Arc and Mutex to be able to lock it between threads

    let sync_dir_clone = Arc::clone(&sync_dir);
    let sync_dir_lock = sync_dir_clone.lock().await;
    (*sync_dir_lock).load_local_files().await; // load local files into the doc
    // subscribe to the doc events, so we can track any changes(for example if other peer adds a new file or changes sth)
    let mut stream = (*sync_dir_lock).doc_subscribe().await?;
    drop(sync_dir_lock);

    // watching for the events in the doc
    let sync_dir_clone_stream = Arc::clone(&sync_dir);
    let stream_task = tokio::spawn(async move { // creating a coroutine using tokio::spawn
        let mut keys: Vec<Vec<u8>> = Vec::new();
        while let Some(item) = stream.next().await { // wait for the item in the stream
            match item {
                Ok(value) => {
                    println!("{:?}", value);
                    match value {
                        LiveEvent::InsertRemote { content_status, entry, .. } => { // remote peer make a change
                            if content_status == ContentStatus::Complete { // if change is already loaded. Otherwise, we wait for the ContentReady
                                let sync_dir_lock = sync_dir_clone_stream.lock().await; // locking syncdir
                                let _ = (*sync_dir_lock).write_new_entry(entry.clone().key().to_vec()).await; // writing new entry to the local fs
                                drop(sync_dir_lock);
                            }else{
                                keys.push(entry.clone().key().to_vec()); // if change is not ready we store key to use it in ContentReady event
                            }
                        }
                        LiveEvent::ContentReady { .. } => { // change loaded (usually invoked after InsertRemote)
                            if !keys.is_empty() { // check if there are any keys left
                                let sync_dir_lock = sync_dir_clone_stream.lock().await;
                                let _ = (*sync_dir_lock).write_new_entry(keys[0].clone()).await; // writing new entry to the local fs
                                drop(sync_dir_lock);
                                keys.remove(0); // removing used key from the list
                            } else {
                                println!("Got ContentReady, but the key is empty");
                            }
                        }
                        LiveEvent::PendingContentReady => { // usually there is nothing in the keys left up to this point, but it is better to make sure. Just in case.
                            for key in &keys {
                                let sync_dir_lock = sync_dir_clone_stream.lock().await;
                                let _ = (*sync_dir_lock).write_new_entry(key.clone()).await;
                                drop(sync_dir_lock);
                            }
                            keys.clear();
                        }
                        _ => {}
                    }
                }
                Err(e) => eprintln!("Error processing stream: {}", e),
            }
        }
    });

    // creating a watcher to watch changes in the local fs
    let sync_dir_clone_watcher = Arc::clone(&sync_dir);
    let watcher_task = tokio::spawn(async move { // creating a coroutine using tokio::spawn
        println!("Start watching...");
        // creating a watcher itself and a channel
        let (tx, rx) = channel();
        let mut watcher = watcher(tx, Duration::from_secs(10)).unwrap();
        watcher.watch(base_path.clone(), RecursiveMode::Recursive).unwrap();

        loop {
            match rx.try_recv() { // waiting got the next item in the stream
                Ok(event) => {
                    println!("Event: {:?}", event);
                    let sync_dir_lock = sync_dir_clone_watcher.lock().await;
                    match event {
                        DebouncedEvent::Create(path) => { // if new entry(file or folder) was created
                            (*sync_dir_lock).get_content_and_load_local_file(path.clone()).await; // write its content and path to the doc
                        }
                        DebouncedEvent::Write(path) => { // if an existing entry(file or folder) was changed
                            (*sync_dir_lock).get_content_and_load_local_file(path.clone()).await; // write its content and path to the doc
                        }
                        DebouncedEvent::Remove(path) => { // if an entry was removed
                            let _ = (*sync_dir_lock).remove_file(path).await; // mark file as deleted in the doc
                        }
                        DebouncedEvent::Rename(src, dst) => { // if to rename an entry or move it to the other directory.
                            // If an entry is moved outside of watching directory, the Remove event will be invoked
                            (*sync_dir_lock).rename_local_file(src, dst).await; // delete an entry in the doc and write it with a new path
                        }
                        _ => {}
                    }
                    drop(sync_dir_lock);
                }
                Err(e) => match e {
                    TryRecvError::Empty => { // if there is nothing new in the stream
                        tokio::time::sleep(Duration::from_millis(watch_rate.clone())).await; // wait watch_rate ms to not block the program with a lot of loop iterations
                    }
                    TryRecvError::Disconnected => {
                        println!("Channel disconnected, stopping watcher.");
                        break;
                    }
                },
            }
        }

        println!("Stop watching");
    });

    // watching for a user input to stop the program
    // It is CRUCIAL to stop it this way because it is required to call *sync_dir_lock).shutdown().await; to persist all the data to local fs
    let mut input = String::new();
    while io::stdin().read_line(&mut input).expect("Failed to read line") > 0 {
        if input.trim() == "q" { // if user entered "q"
            // drop all coroutine
            drop(stream_task);
            drop(watcher_task);
            
            // shutdown node and persist doc to the local fs
            let sync_dir_clone = Arc::clone(&sync_dir);
            let sync_dir_lock = sync_dir_clone.lock().await;
            (*sync_dir_lock).shutdown().await;
            drop(sync_dir_lock);
            break;
        }
    }
    Ok(())
}

