use std::collections::HashMap;
use std::{fs};
use std::fs::File;
use std::io::{Read, Write};
use std::path::{Path, PathBuf};
use std::sync::{Arc};
use futures_lite::{Stream, StreamExt};
use iroh::base::node_addr::AddrInfoOptions;
use iroh::blobs::store;
use iroh::blobs::store::Store;
use iroh::client::docs::{LiveEvent, ShareMode};
use iroh::client::{Doc};
use iroh::docs::{ AuthorId};
use iroh::docs::store::Query;
use iroh::node::Node;
use notify::{Watcher};
use anyhow::{Result};
use bytes::Bytes;
use iroh::base::ticket::Ticket;
use serde::{Deserialize, Serialize};
use tokio::sync::Mutex;

/// struct that represents entity(file or folder) in the local fs
#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Entity {
    pub content: Vec<u8>,
    pub is_deleted: bool,
}

// defining required traits for the Entity
impl Entity {
    fn new(content: Vec<u8>, is_deleted: bool) -> Self{
        Entity{
            content,
            is_deleted
        }
    }
    
    fn from_bytes(bytes: Bytes) -> anyhow::Result<Self> {
        let todo = serde_json::from_slice(&bytes)?;
        Ok(todo)
    }

    fn as_bytes(&self) -> anyhow::Result<Bytes> {
        let buf = serde_json::to_vec(self)?;
        Ok(buf.into())
    }
}

/// struct with all functions to work with Iroh's doc and local fs
pub struct SyncDir {
    base_path: PathBuf, //  sync dir
    store_path: String, // place where to persist data from doc
    node: Node<store::fs::Store>,
    doc: Arc<Mutex<Doc>>,
    author: AuthorId,
}

impl SyncDir{
    pub async fn new(base_path: PathBuf, store_path: String, ticket: Option<String>) -> anyhow::Result<Self>{
        let mut node = iroh::node::Node::persistent(Path::new(&store_path.clone())).await?.spawn().await?;
        let client = node.client();
        
        let doc = match ticket {
            None => { // if there is no ticket specified in the args, create one 
                let d = Arc::new(Mutex::new(client.docs().create().await?));
                let mut doc_lock = d.lock().await;
                let ticket = (*doc_lock).share(ShareMode::Write, AddrInfoOptions::RelayAndAddresses).await?;
                drop(doc_lock);
                println!("{}", ticket);
                d
            }
            Some(ticket) => { // import existing doc from the file system and connect to other peers
                Arc::new(Mutex::new(client.docs().import(ticket.parse()?).await.unwrap())) 
            }
        };
        
        // using a default author
        let author = client.authors().default().await?;
        
        Ok(SyncDir{
            base_path,
            store_path,
            node,
            doc,
            author
        })
    }
    
    /// loading all the files from sync dir to the doc
    pub async fn load_local_files(&self){
        // read all the files/folders from the sync dir
        let directory = self.collect_files_in_dir(self.base_path.clone());
        
        // load all entries to the doc
        for (key, content) in directory {
            let _ = self.load_local_file(key, content).await;
        }
    }

    /// collecting all the files from sync dir
    pub fn collect_files_in_dir<P: AsRef<Path>>(&self, dir: P) -> HashMap<String, Vec<u8>> {
        let mut file_map = HashMap::new();
        let dir = dir.as_ref();

        if dir.is_dir() {
            for entry in fs::read_dir(dir).expect("Failed to read directory") {
                let entry = entry.expect("Failed to read directory entry");
                let path = entry.path();

                if path.is_dir() {
                    // Recursively collect files in subdirectories
                    file_map.insert(path.clone().to_str().unwrap().to_string(), "".into());
                    file_map.extend(self.collect_files_in_dir(path));
                } else if path.is_file() {
                    // Read the file content and insert it into the map
                    let mut file_content = Vec::new();
                    let mut file = fs::File::open(&path).expect("Failed to open file");
                    file.read_to_end(&mut file_content).expect("Failed to read file");

                    file_map.insert(path.clone().to_str().unwrap().to_string(), file_content);
                }
            }
        }

        file_map
    }

    /// writes new entry to the doc if it is different from existing one
    async fn write_to_doc_if_different(&self, key: String, content: Entity) -> anyhow::Result<()>{
        let entity = self.get_last_from_doc(key.clone()).await; // get last entry added to the doc

        match entity {
            Some(entry) => { // if there is already entry with the same key in the doc
                if content.content != entry.content{ // check if contents are different
                    // if so, write new version to the doc
                    let _ = self.set_to_doc(key.clone(), content.content.clone(), false).await;
                }
            },
            None => { // if there is no entry with such key just add new one to the doc
                let _ = self.set_to_doc(key.clone(), content.content.clone(), false).await;
            }
        };
        Ok(())
    }

    /// load file/folder from the local fs to the doc
    pub async fn load_local_file(&self, key: String, content: Vec<u8>) -> anyhow::Result<()> {
        // TODO file might have an empty content as well, so we need to check for it
        // calculate a relative path from sync_dir to the entry
        let relative_path = self.relative_path(PathBuf::from(key.clone()));

        if content.is_empty() { // if the content is empty - it is a folder
            // add 'd' at the begging of the key to indicate it is a folder TODO we might move it to the Entity struct
            let _ = self.write_to_doc_if_different("d".to_owned() + &*relative_path.display().to_string(), Entity::new(vec![0], false)).await;
        } else {
            let content_clone = content.clone();
            let temp = std::str::from_utf8(&content_clone)?;
            println!("load_local_file: {:?}", temp.clone());
            // add 'd' at the begging of the key to indicate it is a file TODO we might move it to the Entity struct
            let _ = self.write_to_doc_if_different("f".to_owned() + &*relative_path.display().to_string(), Entity::new(content.clone(), false)).await;
        }
        
        Ok(())
    }
    
    /// removing file/folder from the local fs
    pub fn remove_local_file(&self, path: PathBuf, is_dir: bool){
        // just removing entry from fs
        if is_dir{
            let _ = fs::remove_dir_all(path);
        }else{
            let _ = fs::remove_file(path);
        }
    }

    /// called when InsertRemote or ContentReady in the doc stream invoked
    pub async fn write_new_entry(&self, key: Vec<u8>) -> anyhow::Result<()> {
        // getting the last entry from the doc by a key
        let entity = self.get_last_from_doc(String::from_utf8(key.clone()).unwrap()).await
            .unwrap_or_else(|| { Entity::new(Vec::new(), false) });

        // deleting 'f' or 'd' that indicate whether it is a file or a folder from the key
        let cloned_key = String::from_utf8(key)?;
        let mut char_indices = cloned_key.char_indices();
        let second_char_start_index = char_indices.nth(1)
            .map(|(index, _)| index)
            .unwrap_or(cloned_key.len());
        let mut entry_name = self.base_path.clone();
        entry_name = entry_name.join(&cloned_key[second_char_start_index..]);
        
        // if an entry marked as deleted in the doc, it should be deleted from the local fs as well
        if entity.is_deleted{
            match cloned_key.chars().next() {
                Some(first_char) if first_char == 'f' => self.remove_local_file(entry_name, false),
                Some(first_char) if first_char == 'd' => self.remove_local_file(entry_name, true),
                Some(_) => {}
                None => {}
            }
            return Ok(()) // no need to check other options
        }
        
        // retrieving content from the entry
        let content_bytes = Bytes::from(entity.content);        

        println!("write_new_entry: {:?}", entry_name);

        match cloned_key.chars().next() {
            Some(first_char) if first_char == 'f' => { // if it is a file
                if let Some(parent) = entry_name.parent() { // creating a missing directory
                    fs::create_dir_all(parent)?;
                }
                let mut file = File::create(entry_name)?; // creating a new file if not already exists
                file.write_all(&*content_bytes)?; // writing new content to the file
            }
            Some(first_char) if first_char == 'd' => { // if it is a folder
                fs::create_dir_all(entry_name)?; // creating new folder
            }
            Some(_) => {}
            None => {}
        }
        
        Ok(())
    }

    /// subscribe to the doc events. The function returns a stream
    pub async fn doc_subscribe(&self) -> Result<impl Stream<Item = Result<LiveEvent>>> {
        let doc_clone = Arc::clone(&self.doc);
        let doc_lock = doc_clone.lock().await;
        let event = doc_lock.subscribe().await;
        drop(doc_lock);
        
        event
    }
    
    /// shuts down a node and persists all the data from the doc to the local fs
    pub async fn shutdown(&self) {
        self.node.clone().shutdown().await.expect("Unable to shutdown node");
    }
    
    /// get last entry from the doc by a key TODO needs to be removed
    pub async fn get_by_key(&self, key: String) -> Option<Entity>{
        self.get_last_from_doc(key.clone()).await
    }

    /// read content from the local file and load it to the doc
    pub async fn get_content_and_load_local_file(&self, path: PathBuf) {
        // retrieving content
        let mut content: Vec<u8> = vec![];
        if path.is_file() {
            let mut file_content = Vec::new();
            let mut file = fs::File::open(&path).expect("Failed to open file");
            file.read_to_end(&mut file_content).expect("Failed to read file");
            content = file_content;
            let str = String::from_utf8(content.clone()).unwrap();
            println!("local content: {str}");
        }
        
        // loading to the doc
        self.load_local_file(path.display().to_string(), content).await.expect("Unable to load local file");
    }

    /// checks if the entry exists in the doc and mark it as deleted
    pub async fn remove_if_exists(&self, key: String) -> anyhow::Result<()> {
        let entry = self.get_last_from_doc(key.clone()).await; // get last entry from the doc
        match entry {
            Some(entity) => {
                // write to the doc an entity with the same content but marked as deleted
                let _ = self.set_to_doc(key.clone(), entity.content, true).await;
            },
            None => {}
        };


        Ok(())
    }
    
    /// checks if the path is a file or a folder and deletes it from the doc
    pub async fn remove_file(&self, path: PathBuf) -> anyhow::Result<()>{
        let relative_path = self.relative_path(path.clone()); // calculate relative path
        
        // considering both options (entry might be a file or a folder)
        let file_path = "f".to_string() + &*relative_path.clone().display().to_string();
        let dir_path = "d".to_string() + &*relative_path.clone().display().to_string();

        // check if file exists
        match self.get_last_from_doc(file_path.clone()).await { 
            Some(entity) if entity.is_deleted != true => {
                let _ = self.remove_if_exists(file_path.clone()).await;
            }
            _ => {} 
        }
        
        // check if directory exists
        match self.get_last_from_doc(dir_path.clone()).await {
            Some(entity) if entity.is_deleted != true => {
                let _ = self.remove_if_exists(dir_path.clone()).await;
            }
            _ => {}
        }
        
        Ok(())
    }
    
    /// calculates relative path from sync dir to any file/folder inside
    pub fn relative_path(&self, path: PathBuf) -> PathBuf{
        match path.strip_prefix(&self.base_path.clone()) {
            Ok(relative_path) => {PathBuf::from(relative_path)}
            Err(_) => {path}
        }
    }
    
    /// write new entry to doc
    pub async fn set_to_doc(&self, key: String, content: Vec<u8>, is_deleted: bool) -> anyhow::Result<()>{
        let doc_clone = Arc::clone(&self.doc);
        let mut doc_lock = doc_clone.lock().await; // locking doc

        let str = String::from_utf8(content.clone()).unwrap(); // for debug
        println!("SET_TO_DOC: {:?}", str);
        
        let entity = Entity::new(content.clone(), is_deleted.clone()); // creating new instance of an entity
        (*doc_lock).set_bytes(self.author, key, entity.as_bytes()?).await?; // writing new entity to the doc
        drop(doc_lock);
        
        Ok(())
    }
    
    /// each author has its own entry in the doc of the same file so this function retrieves the entry that was last added 
    pub async fn get_last_from_doc(&self, key: String) -> Option<Entity> {
        let doc_clone = Arc::clone(&self.doc);
        let mut doc_lock = doc_clone.lock().await;
        // retrieve entry from the doc
        let mut entries = (*doc_lock).get_many(Query::single_latest_per_key().key_exact(key.clone())).await.ok()?; // (returns a stream)
        let entry = entries.next().await; // pick the first element from the stream

        
        let content = match entry { // convert result to Entity and return it
            None => None,
            Some(entry) => {
                match entry { 
                    Ok(entry) => {
                        let content_bytes = entry.content_bytes(&(*doc_lock)).await.ok()?;
                        Some(Entity::from_bytes(content_bytes).unwrap_or(Entity::new(Vec::new(), false)))
                    }
                    Err(_) => {None}
                }
            }
        };
        
        drop(doc_lock);
        
        content
    }
    
    /// renames file/dir inside doc
    pub async fn rename_local_file(&self, src: PathBuf, dst: PathBuf){
        let _ = self.remove_file(src).await; // delete existing file/dir
        let _ = self.get_content_and_load_local_file(dst).await; // write the same file/dir with a new path
    }
}
